#!/bin/bash
set -e

## FUNCTIONS #####################################################
func_make_install () {
   if [ "$UID" -eq 0 -o "$EUID" -eq 0 ]; then
      make install
   else
      sudo make install
   fi
}

func_echo () {
   if [ $VERBOSE == "true" ]; then
      echo "$1"
   fi
}

usage() { 
   echo "Usage: $0 [-v] [-a] [-d <string>] [-w <string>] [-p <string>] [-f <y/n>] [-c <y/n>]"
   echo "  v: disables echo"
   echo "  a: installs all plugins (incompatible with -p)"
   echo "  d: dependencies destination path (either /global/path or relative/path). If not set, will be asked later."
   echo "  w: wolf folder destination path (either /global/path or relative/path). If not set, will be asked later."
   echo "  p: plugins to be installed (incompatible with -a) being a string of 6 chars y/n corresponding to plugins imu, gnss, laser, vision, apriltag, bodydynamics (for example 'ynynnn'). Plugins apriltag and bodydynamics won't be installed if vision and imu plugins are installed, respectively."
   echo "  f: y/n install Falko optional dependency. If not set, will be asked later."
   echo "  c: y/n install CSM optional dependency. If not set, will be asked later."  
}

func_check_yn () {
   if [ "$1" != "y" ] && [ "$1" != "n" ]; then
      echo "options -p, -f, -c should containing only 'y' or 'n'"
      usage
      exit 0
   fi
}

## OPTIONS #####################################################
VERBOSE="true"
INSTALL_PLUGINS=""
INSTALL_IMU="undefined"
INSTALL_GNSS="undefined"
INSTALL_LASER="undefined"
INSTALL_VISION="undefined"
INSTALL_APRILTAG="undefined"
INSTALL_BODYDYNAMICS="undefined"
INSTALL_FALKO="undefined"
INSTALL_CSM="undefined"
DEPS_PATH="undefined"
WOLF_PATH="undefined"

while getopts "vap:d:w:hf:c:" opt; do
   case ${opt} in
      v) # disable verbose
         VERBOSE="false"
         ;;
      a) # install all plugins
      	 if [ "$INSTALL_PLUGINS" != "" ]; then
      	    echo "options -p and -a are incompatible!"
      	    usage
      	    exit 0
      	 fi
         INSTALL_PLUGINS="a"
         INSTALL_IMU="y"
         INSTALL_GNSS="y"
         INSTALL_LASER="y"
         INSTALL_VISION="y"
         INSTALL_APRILTAG="y"
         INSTALL_BODYDYNAMICS="y"
         ;;
      p) # install all plugins
      	 if [ "$INSTALL_PLUGINS" != "" ]; then
      	    echo "options -p and -a are incompatible!"
      	    usage
      	    exit 0
      	 fi
         if [ ${#OPTARG} != 6 ]; then
            echo "option -p should contain 6 characters"
      	    usage
      	    exit 0
      	 fi
         INSTALL_PLUGINS=$OPTARG
         INSTALL_IMU=${INSTALL_PLUGINS:0:1}
         INSTALL_GNSS=${INSTALL_PLUGINS:1:1}
         INSTALL_LASER=${INSTALL_PLUGINS:2:1}
         INSTALL_VISION=${INSTALL_PLUGINS:3:1}
         INSTALL_APRILTAG=${INSTALL_PLUGINS:4:1}
         INSTALL_BODYDYNAMICS=${INSTALL_PLUGINS:5:1}
         func_check_yn $INSTALL_IMU
         func_check_yn $INSTALL_GNSS
         func_check_yn $INSTALL_LASER
         func_check_yn $INSTALL_VISION
         func_check_yn $INSTALL_APRILTAG
         func_check_yn $INSTALL_BODYDYNAMICS
         ;;
      f) # install falko      
         INSTALL_FALKO=$OPTARG
         func_check_yn $INSTALL_FALKO
         ;;
      c) # install csm      
         INSTALL_CSM=$OPTARG
         func_check_yn $INSTALL_CSM
         ;;
      d) # deps path      
         DEPS_PATH=$OPTARG
         ;;
      w) # wolf path       
         WOLF_PATH=$OPTARG
         ;;
      h )
         usage
         exit 0
         ;;
      :)
         echo "Error: -${OPTARG} requires an argument."
         usage
         exit 1
         ;;
      *)
         usage
         exit 1
         ;;
   esac
done

# START #####################################################
CORES=$(nproc)
func_echo "The number of available cores on this machine is $CORES"
RUN_PATH=$PWD

# UBUNTU
if [ "$UID" -eq 0 -o "$EUID" -eq 0 ]; then
   apt install -y lsb-core
else
   sudo apt install -y lsb-core
fi
UBUNTU_DISTRO=$(lsb_release -rs);
if [ $UBUNTU_DISTRO == "16.04" ]; then
   func_echo "Ubuntu 16.04 - OK"
elif [ $UBUNTU_DISTRO == "18.04" ]; then
   func_echo "Ubuntu 18.04 - OK"
elif [ $UBUNTU_DISTRO == "20.04" ]; then
   func_echo "Ubuntu 20.04 - OK"
else
   echo "Non-supported Ubuntu version: $UBUNTU_DISTRO"
   exit 1
fi


# WOLF DEPENDENCIES #####################################################
func_echo "You are in folder $PWD"
if [ $DEPS_PATH == "undefined" ]; then
   echo "Enter path for dependencies (either /global/path or relative/path):"
   read DEPS_PATH
fi
cd $DEPS_PATH
DEPS_PATH=$PWD
func_echo "path dependencies: $DEPS_PATH"

func_echo "Installing dependencies via apt install..."

if [ "$UID" -eq 0 -o "$EUID" -eq 0 ]; then
   #apt install -y git wget unzip dh-autoreconf cmake build-essential libgoogle-glog-dev libgflags-dev libatlas-base-dev libsuitesparse-dev git libboost-all-dev libyaml-cpp-dev wget unzip
   apt install -y git wget unzip cmake build-essential libgoogle-glog-dev libgflags-dev libatlas-base-dev libsuitesparse-dev git libboost-all-dev libyaml-cpp-dev wget unzip
else
   #sudo apt install -y git wget unzip dh-autoreconf cmake build-essential libgoogle-glog-dev libgflags-dev libatlas-base-dev libsuitesparse-dev git libboost-all-dev libyaml-cpp-dev wget unzip
   sudo apt install -y git wget unzip cmake build-essential libgoogle-glog-dev libgflags-dev libatlas-base-dev libsuitesparse-dev git libboost-all-dev libyaml-cpp-dev wget unzip
fi

# gcc version (5.4 for C++14 support)
GCC_REQ=54
GCC_V=$(gcc -v 2>&1 | grep -- 'gcc version' | tr -dc '0-9')
GCC_V=${GCC_V:0:2}
if [ $GCC_V -lt $GCC_REQ ]; then
   echo "The current gcc version ${GCC_V:0:1}.${GCC_V:1:1} does not support C++14. Version ${GCC_REQ:0:1}.${GCC_REQ:1:1} or later should be installed."
   exit 0
fi

# Eigen
# Ubuntu 16.04 - Eigen 3.3.7 required
if [ $UBUNTU_DISTRO == "16.04" ]; then
   func_echo "Installing Eigen 3.3.7 via source..."
   # todo check if installed, check version, check 
   wget https://gitlab.com/libeigen/eigen/-/archive/3.3.7/eigen-3.3.7.zip
   unzip eigen-3.3.7.zip
   rm eigen-3.3.7.zip
   cd eigen-3.3.7
   mkdir build
   cd build
   cmake ..
   func_make_install
else
   func_echo "Installing Eigen 3 via apt install"
   apt install -y libeigen3-dev
fi 

# ceres
func_echo "Installing ceres 2.0 via source..."
cd $DEPS_PATH
git clone --depth 1 -b 2.0.0 https://ceres-solver.googlesource.com/ceres-solver
cd ceres-solver
mkdir -p build && cd build
cmake -DCMAKE_BUILD_TYPE=Release ..
make -j$CORES
make test
func_make_install

# spdlog
func_echo "Installing spdlog 0.17 via source..."
cd $DEPS_PATH
git clone --depth 1 -b v0.17.0 https://github.com/gabime/spdlog.git
cd spdlog
mkdir -p build && cd build
cmake -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="${CMAKE_CXX_FLAGS} -fPIC" ..
make -j$CORES
func_make_install


# WOLF #####################################################
cd $RUN_PATH
func_echo "You are in folder $PWD"
if [ $WOLF_PATH == "undefined" ]; then
   echo "Enter path for wolf folder (either /global/path or relative/path):"
   read WOLF_PATH
fi
cd $WOLF_PATH
mkdir wolf
cd wolf
WOLF_PATH=$PWD
func_echo "wolf folder path: $WOLF_PATH"


# CORE -----------------------------------------------------
func_echo "Cloning wolf core..."
cd $WOLF_PATH
git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/wolf.git
cd wolf

func_echo "Compiling wolf core..."
mkdir -p build && cd build
cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_DEMOS=ON -DBUILD_TESTS=ON ..
make -j$CORES

func_echo "Installing wolf core..."
func_make_install

func_echo "Testing wolf core..."
ctest -j$CORES


# IMU --------------------------------------------------
if [ $INSTALL_IMU == "undefined" ]; then
   echo "Do you want to download and install plugin imu? (y/n)"
   read INSTALL_IMU
fi
if [ $INSTALL_IMU == "y" ]; then

   func_echo "Cloning plugin imu..."
   cd $WOLF_PATH
   git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/imu.git
   cd imu
   
   func_echo "Compiling plugin imu..."
   mkdir -p build && cd build
   cmake .. -DCMAKE_BUILD_TYPW=Release
   make -j$CORES

   func_echo "Installing plugin imu..."
   func_make_install

   func_echo "Testing plugin imu..."
   ctest -j$CORES
else
   func_echo "Skipping plugin imu."
fi

# GNSS --------------------------------------------------
if [ $INSTALL_GNSS == "undefined" ]; then
   echo "Do you want to download and install plugin gnss? (y/n)"
   read INSTALL_GNSS
fi
if [ $INSTALL_GNSS == "y" ]; then

   func_echo "Installing plugin gnss dependencies..."
   cd $DEPS_PATH
   git clone https://gitlab.iri.upc.edu/mobile_robotics/gauss_project/gnss_utils.git
   cd gnss_utils
   git submodule update --init
   mkdir -p build && cd build
   cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_DEMOS=ON -DBUILD_TESTS=ON ..
   make -j$CORES
   ctest -j$CORES
   func_make_install

   func_echo "Cloning plugin gnss..."
   cd $WOLF_PATH
   git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/gnss.git
   cd gnss

   func_echo "Compiling plugin gnss..."
   mkdir -p build && cd build
   cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_DEMOS=ON -DBUILD_TESTS=ON ..
   make -j$CORES

   func_echo "Installing plugin gnss..."
   func_make_install

   func_echo "Testing plugin gnss..."
   ctest -j$CORES
else
   func_echo "Skipping plugin gnss."
fi


# LASER --------------------------------------------------
if [ $INSTALL_LASER == "undefined" ]; then
   echo "Do you want to download and install plugin laser? (y/n)"
   read INSTALL_LASER
fi
if [ $INSTALL_LASER == "y" ]; then

   func_echo "Installing plugin laser dependencies..."

   # CSM
   if [ $INSTALL_CSM == "undefined" ]; then
      echo "Do you want to install CSM to enable ICP processors? (y/n)"
      read INSTALL_CSM
   fi
   if [ $INSTALL_CSM == "y" ]; then
      if [ "$UID" -eq 0 -o "$EUID" -eq 0 ]; then
         apt install -y libgsl-dev
      else
         sudo apt install -y libgsl-dev
      fi
      cd $DEPS_PATH
      git clone https://gitlab.iri.upc.edu/labrobotica/algorithms/csm.git
      cd csm
      mkdir -p build && cd build
      cmake -DCMAKE_BUILD_TYPE=Release ..
      make -j$CORES
      func_make_install
   else 
      func_echo "Skipping CSM."
   fi

   # FALKO
   if [ $INSTALL_FALKO == "undefined" ]; then
      echo "Do you want to install FALKO to enable Falko loop closure processors? (y/n)"
      read INSTALL_FALKO
   fi
   if [ $INSTALL_FALKO == "y" ]; then
      cd $DEPS_PATH
      git clone https://gitlab.iri.upc.edu/labrobotica/algorithms/falkolib.git
      cd falkolib
      mkdir -p build && cd build
      cmake -DCMAKE_BUILD_TYPE=Release ..
      make -j$CORES
      func_make_install
   else 
      func_echo "Skipping Falko."
   fi

   # LASER_SCAN_UTILS
   func_echo "Installing laser_scan_utils (required)..."
   cd $DEPS_PATH
   ldconfig
   git clone https://gitlab.iri.upc.edu/labrobotica/algorithms/laser_scan_utils.git
   cd laser_scan_utils
   mkdir -p build && cd build
   cmake -DCMAKE_BUILD_TYPE=Release ..
   make -j$CORES
   ctest -j$CORES
   func_make_install

   # PLUGIN
   func_echo "Cloning plugin laser..."
   cd $WOLF_PATH
   ldconfig
   git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/laser.git
   cd laser

   func_echo "Compiling plugin laser..."
   mkdir -p build && cd build
   cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_DEMOS=ON -DBUILD_TESTS=ON ..
   make -j$CORES

   func_echo "Installing plugin laser..."
   func_make_install

   func_echo "Testing plugin laser..."
   ldconfig
   ctest -j$CORES
else
   func_echo "Skipping plugin laser."
fi

# VISION --------------------------------------------------
if [ $INSTALL_VISION == "undefined" ]; then
   echo "Do you want to download and install plugin vision (requires to install opencv 3.3.0)? (y/n)"
   read INSTALL_VISION
fi
if [ $INSTALL_VISION == "y" ]; then

   func_echo "Installing plugin vision dependencies..."
   cd $DEPS_PATH

   #install opencv
   wget https://github.com/opencv/opencv/archive/3.3.0.zip -O temp.zip
   unzip temp.zip
   rm temp.zip
   cd opencv-3.3.0
   git clone --depth 1 -b 3.3.0 https://github.com/opencv/opencv_contrib.git
   mkdir build
   cd build
   cmake -D CMAKE_BUILD_TYPE=Release -D CMAKE_INSTALL_PREFIX=/usr/local -D CMAKE_CXX_FLAGS="-fPIC -std=c++14" -D OPENCV_EXTRA_MODULES_PATH=../opencv_contrib/modules ..
   make -j$CORES
   func_make_install

   func_echo "Cloning plugin vision..."
   cd $WOLF_PATH
   git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/vision.git
   cd vision

   func_echo "Compiling plugin vision..."
   mkdir -p build && cd build
   cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_DEMOS=ON -DBUILD_TESTS=ON ..
   make -j$CORES

   func_echo "Installing plugin vision..."
   func_make_install

   func_echo "Testing plugin vision..."
   ctest -j$CORES
else
   func_echo "Skipping plugin vision."
fi

# APRILTAG --------------------------------------------------
if [ $INSTALL_VISION == "y" ]; then
   if [ $INSTALL_APRILTAG == "undefined" ]; then
      echo "Do you want to download and install plugin apriltag? (y/n)"
      read INSTALL_APRILTAG
   fi
   if [ $INSTALL_APRILTAG == "y" ]; then

      func_echo "Installing plugin apriltag dependencies..."
      cd $DEPS_PATH
      git clone https://github.com/AprilRobotics/apriltag apriltaglib
      cd apriltaglib
      mkdir -p build && cd build
      cmake -DCMAKE_BUILD_TYPE=Release ..
      make -j$CORES
      ctest -j$CORES
      func_make_install

      func_echo "Cloning plugin apriltag..."
      cd $WOLF_PATH
      git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/apriltag.git
      cd apriltag

      func_echo "Compiling plugin apriltag..."
      mkdir -p build && cd build
      cmake -DCMAKE_BUILD_TYPE=Release -DBUILD_DEMOS=ON -DBUILD_TESTS=ON ..
      make -j$CORES

      func_echo "Installing plugin apriltag..."
      func_make_install

      func_echo "Testing plugin apriltag..."
      ctest -j$CORES
   else
      func_echo "Skipping plugin apriltag."
   fi
else
   func_echo "Skipping plugin apriltag since plugin vision was not installed."
fi

# BODYDYNAMICS --------------------------------------------------
if [ $INSTALL_IMU == "y" ]; then
   if [ $INSTALL_BODYDYNAMICS == "undefined" ]; then
      echo "Do you want to download and install plugin bodydynamics? (y/n)"
      read INSTALL_BODYDYNAMICS
   fi
   if [ $INSTALL_BODYDYNAMICS == "y" ]; then

      func_echo "Cloning plugin bodydynamics..."
      cd $WOLF_PATH
      git clone -b master https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/bodydynamics.git
      cd bodydynamics

      func_echo "Compiling plugin bodydynamics..."
      mkdir -p build && cd build
      cmake .. -DCMAKE_BUILD_TYPW=Release
      make -j$CORES

      func_echo "Installing plugin bodydynamics..."
      func_make_install

      func_echo "Testing plugin bodydynamics..."
      ctest -j$CORES
   else
      func_echo "Skipping plugin bodydynamics."
   fi
else
   func_echo "Skipping plugin bodydynamics since plugin imu was not installed."
fi

func_echo "Done! Enjoy!"